import torch
from torch_geometric.data import InMemoryDataset, Data, Dataset
from torch_geometric.utils import to_undirected
from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle
import random
import numpy as np
import pickle
import pandas as pd
import os.path as osp
import itertools
import os
from icecream import ic


class Dataset(InMemoryDataset):
    def __init__(self, root, dataset, rating_file, sep, args, transform=None, pre_transform=None):

        # '../data/'
        self.path = root
        # 'book-crossing'
        self.dataset = dataset
        # 'implicit_ratings.csv'
        self.rating_file = rating_file
        # ','
        self.sep = sep
        self.store_backup = True
        self.args = args

        super(Dataset, self).__init__(root, transform, pre_transform)
        # processed_file_names来自下面的@property自定义属性
        # self.processed_paths: ['../data/processed/book-crossing/book-crossing.dataset', '../data/processed/book-crossing/book-crossing.statinfo']
        # self.data: Data(x=[11266278, 1], edge_index=[2, 78302822], edge_attr=[48682656, 2], y=[1050834])
        # self.slices: {'x': tensor(1050835), 'edge_index': tensor(1050835), 'edge_attr': tensor(1050835), 'y': tensor(1050835)}
        self.data, self.slices = torch.load(self.processed_paths[0])
        # {'data_num': 1050834, 'feature_num': 101285, 'train_test_split_index': [630500, 840667]}
        self.stat_info = torch.load(self.processed_paths[1])
        # 1050834
        self.data_num = self.stat_info['data_num']
        # 101285
        self.feature_num = self.stat_info['feature_num']

    @property
    def raw_file_names(self):
        return ['{}{}/user_dict.pkl'.format(self.path, self.dataset),
                '{}{}/item_dict.pkl'.format(self.path, self.dataset),
                '{}{}/feature_dict.pkl'.format(self.path, self.dataset),
                '{}{}/{}'.format(self.path, self.dataset, self.rating_file)]

    @property
    def processed_file_names(self):
        return ['{}/{}.dataset'.format(self.dataset, self.dataset),
                '{}/{}.statinfo'.format(self.dataset, self.dataset)]

    def download(self):
        # Download to `self.raw_dir`.
        pass

    # DataFrame: (630500, 3)
    def data_2_graphs(self, ratings_df, dataset='train'):
        graphs = []
        processed_graphs = 0
        # 630500
        num_graphs = ratings_df.shape[0]
        # 630
        one_per = int(num_graphs / 1000)
        percent = 0.0
        for i in range(len(ratings_df)):
            if processed_graphs % one_per == 0:
                print(f"Processing [{dataset}]: {percent / 10.0}%, {processed_graphs}/{num_graphs}", end="\r")
                percent += 1
            processed_graphs += 1
            # Series: (0, 51883) (1, '3608934227') (2, 0.0)
            line = ratings_df.iloc[i]
            # '51883'
            user_index = self.user_key_type(line[0])
            # '0553374966'
            item_index = self.item_key_type(line[1])
            # 0
            rating = int(line[2])

            if item_index not in self.item_dict or user_index not in self.user_dict:
                error_num += 1
                continue

            # 933
            user_id = self.user_dict[user_index]['name']
            # 64984
            item_id = self.item_dict[item_index]['title']

            # [1, 16]
            user_attr_list = self.user_dict[user_index]['attribute']
            # [11581, 55114, 16138, 5686, 64985, 58689, 11582, 4977, 64986]
            item_attr_list = self.item_dict[item_index]['attribute']

            # [933, 1, 16]
            user_list = [user_id] + user_attr_list
            # [41338, 17321, 9929, 29902, 13602, 10083, 15884, 8086, 4964, 5174]
            item_list = [item_id] + item_attr_list

            graph = self.construct_graphs(user_list, item_list, rating)

            graphs.append(graph)

        print()

        return graphs

    def read_data(self):
        # 观察到attribute似乎都是两个，不像item的那个是属性个数是变化的
        # dict(4873) {
        #   '243': {'name': 0, 'attribute': [1, 2]},
        #   '254': {'name': 3, 'attribute': [1, 4]},...
        # }
        self.user_dict = pickle.load(open(self.userfile, 'rb'))
        # dict(53168) {
        #   '0002005018': {'title': 4960, 'attribute': [4961, 4962, 4963, 4964, 4965]},
        #   '0374157065': {'title': 4966, 'attribute': [4967, 4968, 4969, 4970, 4971, 4972, 4973, 4974, 4975, 4976, 4977, 4978]},...
        # }
        self.item_dict = pickle.load(open(self.itemfile, 'rb'))
        # <class 'str'>
        self.user_key_type = type(list(self.user_dict.keys())[0])
        # <class 'str'>
        self.item_key_type = type(list(self.item_dict.keys())[0])
        # dict(101285) {'user_243': 0, 'usa': 1, 'age_unknown': 2, 'user_254': 3, 'age_15': 4, ..., 'age_12': 8, ..., 'france': 24, ...}
        feature_dict = pickle.load(open(self.featurefile, 'rb'))

        data = []
        error_num = 0

        # DataFrame: (1050834, 3)
        ratings_df = pd.read_csv(self.ratingfile, sep=self.sep, header=None)
        # DataFrame: (630500, 3)    DataFrame: (420334, 3)
        train_df, test_df = train_test_split(ratings_df, test_size=0.4, random_state=self.args.random_seed, stratify=ratings_df[[0, 2]])
        # DataFrame: (210167, 3)    DataFrame: (210167, 3)
        test_df, valid_df = train_test_split(test_df, test_size=0.5, random_state=self.args.random_seed, stratify=test_df[[0, 2]])

        # store a backup of train/valid/test dataframe
        if self.store_backup:
            backup_path = f"{self.path}{self.dataset}/split_data_backup/"
            if not os.path.exists(backup_path):
                os.mkdir(backup_path)

            train_df.to_csv(f'{backup_path}train_data.csv', index=False)
            valid_df.to_csv(f'{backup_path}valid_data.csv', index=False)
            test_df.to_csv(f'{backup_path}test_data.csv', index=False)

        print('(Only run at the first time training the dataset)')
        train_graphs = self.data_2_graphs(train_df, dataset='train')
        valid_graphs = self.data_2_graphs(valid_df, dataset='valid')
        test_graphs = self.data_2_graphs(test_df, dataset='test')

        graphs = train_graphs + valid_graphs + test_graphs

        stat_info = {}
        stat_info['data_num'] = len(graphs)
        stat_info['feature_num'] = len(feature_dict)
        stat_info['train_test_split_index'] = [len(train_graphs), len(train_graphs) + len(valid_graphs)]

        print('error number of data:', error_num)
        return graphs, stat_info

    # [2093, 1, 2]   [41338, 17321, 9929, 29902, 13602, 10083, 15884, 8086, 4964, 5174]   0
    def construct_graphs(self, user_list, item_list, rating):

        # 3
        u_n = len(user_list)  # user node number
        # 10
        i_n = len(item_list)  # item node number

        # construct full inner edge
        # [[0, 0, 0, 1, 1, 2],
        #  [0, 1, 2, 1, 2, 2]]
        inner_edge_index = [[], []]
        for i in range(u_n):
            for j in range(i, u_n):
                inner_edge_index[0].append(i)
                inner_edge_index[1].append(j)

        # 上面是构建user属性内部间的全连接子图，这里是item内部的全连接子图
        for i in range(u_n, u_n + i_n):
            for j in range(i, u_n + i_n):
                inner_edge_index[0].append(i)
                inner_edge_index[1].append(j)

        # construct outer edge
        # [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2],
        #  [3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]]
        outer_edge_index = [[], []]
        for i in range(u_n):
            for j in range(i_n):
                outer_edge_index[0].append(i)
                outer_edge_index[1].append(u_n + j)

        # construct graph
        # torch.Size([2, 61])
        inner_edge_index = torch.LongTensor(inner_edge_index)
        # torch.Size([2, 109])
        inner_edge_index = to_undirected(inner_edge_index)
        # torch.Size([2, 30])
        outer_edge_index = torch.LongTensor(outer_edge_index)
        # torch.Size([2, 60])
        outer_edge_index = to_undirected(outer_edge_index)
        # Data(x=[13, 1], edge_index=[2, 109], edge_attr=[60, 2], y=[1])
        graph = self.construct_graph(user_list + item_list, inner_edge_index, outer_edge_index, rating)

        return graph

    # edge_index_inner: torch.Size([2, 109])    edge_index_outer: torch.Size([2, 60])
    def construct_graph(self, node_list, edge_index_inner, edge_index_outer, rating):
        # torch.Size([13, 1])
        x = torch.LongTensor(node_list).unsqueeze(1)
        # tensor([0.])
        rating = torch.FloatTensor([rating])
        # 构造每个样本
        return Data(x=x, edge_index=edge_index_inner, edge_attr=torch.transpose(edge_index_outer, 0, 1), y=rating)

    def process(self):
        # '../data/book-crossing/user_dict.pkl'
        self.userfile = self.raw_file_names[0]
        # '../data/book-crossing/item_dict.pkl'
        self.itemfile = self.raw_file_names[1]
        # '../data/book-crossing/feature_dict.pkl'
        self.featurefile = self.raw_file_names[2]
        # '../data/book-crossing/implicit_ratings.csv'
        self.ratingfile = self.raw_file_names[3]
        graphs, stat_info = self.read_data()
        # check whether foler path exist
        if not os.path.exists(f"{self.path}processed/{self.dataset}"):
            os.mkdir(f"{self.path}processed/{self.dataset}")

        # collate方法是InMemoryDataset 类所特有的，可查看官网注释：https://pytorch-geometric.readthedocs.io/en/latest/modules/data.html#torch_geometric.data.Dataset
        data, slices = self.collate(graphs)
        torch.save((data, slices), self.processed_paths[0])

        torch.save(stat_info, self.processed_paths[1])

    def feature_N(self):
        return self.feature_num

    def data_N(self):
        return self.data_num
